import pkg_resources
from sinagot.models import Workspace
from .script import ScriptTemplate


__all__ = ["Workspace", "ScriptTemplate"]

__version__ = pkg_resources.get_distribution("sinagot").version



# coding=utf-8

import sys
import logging
from sinagot.utils import (
    record_log_file_path,
    LOG_ORIGIN,
    LOG_RECORD_ID,
    LOG_STEP_LABEL,
)


LOG_FORMAT_UNIT = "%(asctime)s | %({})s : %(message)s".format(LOG_STEP_LABEL)

LOG_FORMAT = "%(asctime)s | %(task)s-%(modality)s-%({})s : %(message)s".format(
    LOG_STEP_LABEL
)


def logger_factory(config, is_unit=False):
    """Create logger with config info"""

    if "log" in config:
        config = config["log"]
    else:
        config = {
            "name": "sinagot",
            "format": "%(asctime)s : %(message)s",
            "level": "INFO",
        }

    logger = logging.getLogger(config["name"])

    handlers = list(logger.handlers)
    for hdlr in handlers:
        logger.removeHandler(hdlr)

    handler = logging.StreamHandler(sys.stdout)
    if "format" in config:
        formatter = logging.Formatter(config["format"])
        handler.setFormatter(formatter)
    handler.addFilter(filter_no_script)
    logger.addHandler(handler)
    if "level" in config:
        logger.setLevel(config["level"])
    logger.debug("log initialized")

    script_handler = logging.StreamHandler(sys.stdout)
    if is_unit:
        log_format = LOG_FORMAT_UNIT
    else:
        log_format = LOG_FORMAT
    script_handler.setFormatter(
        logging.Formatter("%({})s | ".format(LOG_RECORD_ID) + log_format)
    )
    script_handler.addFilter(filter_is_script)
    logger.addHandler(script_handler)

    return logger


def get_log_origin_script(record):
    return record.__dict__.get(LOG_ORIGIN, "script")


def filter_no_script(record):
    return get_log_origin_script(record) != "script"


def filter_is_script(record):
    return get_log_origin_script(record) == "script"




# coding=utf-8


class Model:
    """Base class for every model except scripts"""

    _REPR_ATTRIBUTES = []
    """List of instance attributes to display in __repr__"""
    workspace = None
    """Master workspace"""
    config = None
    """Config dic of the workspace"""
    logger = None
    """Logger object"""

    def __init__(self, workspace):
        self.workspace = workspace
        self.config = workspace.config
        self.logger = workspace.logger

    def __repr__(self):
        attributes = self._get_repr_attributes()
        if self._REPR_ATTRIBUTES:
            attributes = " | " + ", ".join(
                [
                    "{label}: {value}".format(label=label, value=getattr(self, label))
                    for label in self._get_repr_attributes()
                    if hasattr(self, label)
                ]
            )
        else:
            attributes = ""
        return "<{} instance{}>".format(self.__class__.__name__, attributes)

    @property
    def is_unit(self) -> bool:
        return True

    def _get_repr_attributes(self):
        return self._REPR_ATTRIBUTES



from sinagot.utils import get_script


class RunManager:
    """Manage multiple run in sequential or parallel mode"""

    def __init__(self, workspace):
        self.workspace = workspace

    def run(self, model, **run_opts):

        if model._MODEL_TYPE == "record":
            records = [model]
        else:
            records = model.all()

        return self._run(records, run_opts)

    @staticmethod
    def record_structure(record):
        return record._structure_status(("record_id", "task", "modality", "step_label"))

    def _run(self, records, run_opts):

        for record in records:
            shape = self.record_structure(record)
            for task in shape.values():
                for modality in task.values():
                    for step in modality:
                        step_kw = {
                            key: step[key]
                            for key in ("record_id", "task", "modality", "step_label")
                        }
                        run_step_factory(self.workspace, **step_kw, run_opts=run_opts)()

    def close(self):
        pass


def run_step_factory(workspace, record_id, task, modality, step_label, run_opts):
    script = get_script(
        workspace=workspace,
        record_id=record_id,
        task=task,
        modality=modality,
        step_label=step_label,
    )

    run_opts_ = run_opts.copy()
    step_label_filter = run_opts_.pop("step_label", None)
    to_run = step_label_filter in [step_label, None]

    def func(*args):
        if to_run:
            script._run(**run_opts_)

    func.__name__ = step_label
    return func



# coding=utf-8

from importlib import import_module
from typing import Optional, Generator
import pandas as pd
from sinagot.utils import get_module, get_plugin_modules
from sinagot.models import Model


class Scope(Model):
    """
    A Scope instance can :

    - Handle task and modality subscopes.

    - Handle StepCollection with `.steps` attribute.

    Note:
        Scope is base class for [RecordCollection](record_collection.md) and [Record](record.md).
    """

    _REPR_ATTRIBUTES = ["task", "modality"]
    _MODEL_TYPE = None
    _subscope_class = None
    _tasks = []
    task = None
    _modalities = []
    modality = None

    def __init__(
        self,
        workspace: "Workspace",
        task: Optional[str] = None,
        modality: Optional[str] = None,
    ):
        """
        Args:
            workspace: Root Workspace.
            task: Task of the scope.
            modality: Modality of the scope.
        """

        if task is not None:
            self.task: Optional[str] = task
            """Task of the scope. If `None`, the scope represents all available tasks."""
        if modality is not None:
            self.modality: Optional[str] = modality
            """Modality of the scope. If `None`, the scope represents all available modalities."""
        if self._subscope_class is None:
            self._subscope_class = self.__class__
        super().__init__(workspace)

    def _get_repr_attributes(self):
        attributes = super()._get_repr_attributes()
        if not self.workspace.is_unit_mode:
            for attribute in ["task", "modality"]:
                if attribute not in attributes:
                    attributes.append(attribute)
        return attributes

    @classmethod
    def _set_subscopes(cls, workspace):
        """Create subscope tasks and modalities attributes
        Run once at workspace initialization"""

        SUBSCOPES = (
            ("task", "tasks"),
            ("modality", "modalities"),
        )

        for subscope, collection in SUBSCOPES:
            if collection in workspace.config:
                setattr(
                    cls,
                    "_{}".format(collection),
                    [
                        cls._add_subscope(
                            subscope=subscope, value=value, workspace=workspace
                        )
                        for value in workspace.config[collection].keys()
                    ],
                )

    @classmethod
    def _add_subscope(cls, subscope, value, workspace):
        """Add subscope to self.__class__ as a property
        that call an instance of local defined subclass"""

        _subscope_class = cls._search_custom_subscope_in_config(
            subscope, value, workspace
        )
        subscope_access = property(
            cls._property_factory(_subscope_class, subscope, value)
        )
        setattr(
            cls, value, subscope_access,
        )
        return value

    @classmethod
    def _search_custom_subscope_in_config(cls, subscope, value, workspace):
        model_type = cls._MODEL_TYPE
        if subscope == "modality":

            sub_config = workspace.config
            for key in ("modalities", value, "models"):
                sub_config = sub_config.get(key, {})

            custom_class = cls._get_custom_class(workspace, value)
            if custom_class:
                return custom_class

            plugin_model = sub_config.get("plugin")
            if plugin_model:
                return get_plugin_modules(plugin_model)[model_type]
        return cls

    @classmethod
    def _get_custom_class(cls, workspace, value):
        model_type = cls._MODEL_TYPE
        base_class_name = model_type.title().replace("_", "")
        try:
            custom_class = get_module(
                workspace, base_class_name, value, "models", model_type
            )
        except FileNotFoundError:
            return None
        custom_class.__name__ = value.title() + base_class_name
        if model_type == "record_collection":
            record_class = custom_class._record_class._get_custom_class(
                workspace, value
            )
            if record_class:
                custom_class._record_class = record_class
        return custom_class

    @staticmethod
    def _property_factory(_subscope_class, subscope, value):
        def get(self):
            if self._is_valid_subscope(subscope, value):
                kwargs = {
                    "workspace": self.workspace,
                    "task": self.task,
                    "modality": self.modality,
                }
                kwargs[subscope] = value
                if self._MODEL_TYPE == "record":
                    kwargs["record_id"] = self.id
                return _subscope_class(**kwargs)
            else:
                raise AttributeError(
                    "{} {} is not valid for {}".format(subscope, value, self)
                )

        return get

    def _is_valid_subscope(self, subscope, value):
        if getattr(self, subscope) is not None:
            return False
        if subscope == "task":
            return self.modality in self.workspace.config["tasks"][value][
                "modalities"
            ] + [None]
        elif subscope == "modality":
            return (
                self.task is None
                or value in self.workspace.config["tasks"][self.task]["modalities"]
            )
        else:
            return False

    @property
    def is_unit(self) -> bool:
        """Check if has both task and modality not `None`."""

        return not (self.task is None or self.modality is None)

    def get_subscope(
        self, task: Optional[str] = None, modality: Optional[str] = None
    ) -> "Scope":
        """Get subscope for given task and/or modality

        Args:
            task : the task target
            modality : the modality target

        Returns:
            [type]: [description]
        """
        result = self
        if task:
            result = getattr(result, task)
        if modality:
            result = getattr(result, modality)
        return result

    def iter_units(self) -> Generator["Scope", None, None]:
        """
        Generate each 'unit' subscopes of the current scope,
        i.e. subscope with specific task and modality

        Returns:
            Subscope with a task attribute.
        """

        if self.task is None:
            tasks = self.iter_tasks()
        else:
            tasks = [self]
        for task in tasks:
            if task.modality is None:
                modalities = task.iter_modalities()
            else:
                modalities = [task]
            for modality in modalities:
                yield modality

    def iter_tasks(self) -> Generator["Scope", None, None]:
        """
        Generate subscopes of the current scope for each of its task.

        Returns:
            Subscope with a task attribute.
        """

        for task in self._tasks:
            if self._is_valid_subscope("task", task):
                model = getattr(self, task)
                yield model

    def iter_modalities(self) -> Generator["Scope", None, None]:
        """
        Generate subscopes of the current scope for each of its modalities.

        Returns:
            Subscope with a modality attribute.
        """

        for modality in self._modalities:
            if self._is_valid_subscope("modality", modality):
                yield getattr(self, modality)

    def count_detail(self, *args, **kwargs):

        if self.is_unit:
            df = self._count_detail_unit(*args, **kwargs)
        else:
            df = pd.concat(
                [unit.count_detail(*args, **kwargs) for unit in self.iter_units()]
            )
        df.reset_index(drop=True, inplace=True)
        return df

    def _count_detail_unit(self):
        return pd.DataFrame()



# coding=utf-8

import inspect
from typing import Optional
from pathlib import Path
import pandas as pd
from sinagot.models import Model
from sinagot.utils import StepStatus, LOG_STEP_LABEL, LOG_STEP_STATUS, get_module
from sinagot.models.exceptions import NotFoundError, NoModalityError


class Step(Model):

    _REPR_ATTRIBUTES = ["scope", "task", "modality", "label"]
    _MODEL_TYPE = "step"
    task = None
    modality = None

    def __init__(self, script, model):

        self.model = model
        super().__init__(model.workspace)

        if inspect.isclass(script):
            script_class = script
            label = script.__class__.__name__
        elif isinstance(script, str):
            try:
                script_class = get_module(
                    self.workspace, "Script", *self._scripts_folder, script
                )
            except FileNotFoundError as ex:
                raise NotFoundError from ex
            label = script
        else:
            raise AttributeError("Type {} is not valid for script".format(type(script)))
        self.label = label

        self.script_class = script_class
        self.script = script_class(
            data_path=self.workspace.data_path,
            record_id=self.id,
            task=self.task,
            logger_namespace=self.logger.name,
        )

    @property
    def _scripts_folder(self):
        return []

    @property
    def id(self):
        if self.model._MODEL_TYPE == "record":
            return self.model.id

    def status(self) -> int:
        """Get status code.

        Returns:
            sinagot.utils.StepStatus value
        """
        if self._script_path_exists("output"):
            return StepStatus.DONE
        try:
            logs = self.logs()
            status = logs[logs[LOG_STEP_STATUS].notna()].iloc[0][LOG_STEP_STATUS]
        except:
            status = None
        if status in [StepStatus.PROCESSING, StepStatus.ERROR]:
            return status
        if self._script_path_exists("input"):
            return StepStatus.DATA_READY
        else:
            return StepStatus.INIT

    def _script_path_exists(self, position: str) -> bool:
        path = getattr(self.script.path, position)
        if isinstance(path, Path):
            return path.exists()
        if isinstance(path, dict):
            return all([p.exists() for p in path.values()])

    def run(
        self,
        force: Optional[bool] = False,
        ignore_missing: Optional[bool] = False,
        debug: Optional[bool] = False,
    ):
        """
        Run the step script.
        
        Args:
            force: Force run and overwrites result file(s) if already exist(s).
        """
        self.model.steps.run(
            step_label=self.label,
            force=force,
            ignore_missing=ignore_missing,
            debug=debug,
        )

    def logs(self) -> pd.DataFrame:
        """
        Returns:
            Logs history.
        """
        try:
            return self.model.logs().query(
                "{} == '{}'".format(LOG_STEP_LABEL, self.label)
            )
        except:
            return pd.DataFrame()


class ScopedStep(Step):
    def __init__(self, script, model):
        if not model.modality:
            raise NoModalityError
        self.task = model.task
        self.modality = model.modality
        super().__init__(script, model)

    @property
    def _scripts_folder(self):
        return [self.modality]



# coding=utf-8

from pathlib import Path
import pandas as pd
from sinagot.utils import get_plugin_modules
from sinagot.models import Scope, Record, RecordCollection


class PandasPath(Scope):
    @property
    def _data_raw_path(self):
        return self.config["modalities"][self.modality]["models"]["data_path"]


class PandasRecord(Record, PandasPath):
    def _get_data(self):
        if Path(self._data_path).exists():
            try:
                return pd.read_json(self._data_path, typ="series")
            except:
                return None

    def _set_data(self, data):
        if isinstance(data, dict):
            data = pd.Series(data)
            print("data", data)
        if isinstance(data, pd.Series):
            data.T.to_json(self._data_path)

    data = property(_get_data, _set_data)

    @property
    def _data_path(self):
        data_path = self._data_raw_path
        data_path = Path(self.workspace.data_path).joinpath(*data_path)
        data_path = str(data_path).format(id=self.id)
        Path(data_path).parent.mkdir(exist_ok=True)
        return data_path

    def _unit_count_raw_data(self):
        if self.data is None:
            return 0
        return 1


class PandasRecordCollection(RecordCollection, PandasPath):

    ID_FIELD = "record_id"
    DATE_FIELDS = ["date_of_birth", "date_of_recording"]

    _record_class = PandasRecord
    is_unit = True

    @property
    def data(self):
        return pd.DataFrame({rec.id: rec.data for rec in self.all()}).T

    def _iter_path(self):
        return self._data_raw_path


MODELS = {"record": PandasRecord, "record_collection": PandasRecordCollection}



# coding=utf-8

import json
from collections import defaultdict
from io import StringIO
import pandas as pd
from sinagot.models import ModelWithStepCollection, Scope
from sinagot.utils import (
    record_log_file_path,
    LOG_STEP_LABEL,
    LOG_STEP_STATUS,
)
from sinagot.models.exceptions import NotUnitError


class RecordUnit(ModelWithStepCollection):

    _REPR_ATTRIBUTES = ["id"]
    _MODEL_TYPE = "record"

    def __init__(self, workspace: "Workspace", record_id: str, *args, **kwargs):
        """
        Args:
            workspace: Root Workspace.
            record_id: ID of the record.
        """

        # TODO: raise exception if id not match regex in config
        self.id = record_id
        super().__init__(workspace, *args, **kwargs)
        self.steps: "StepCollection" = self.set_step_collection()
        """Collection of steps."""

    def _structure_status(self, columns):

        df = self.steps.status()

        def status_default_dict():
            def dict_to_list():
                return defaultdict(list)

            return defaultdict(dict_to_list)

        def get_status_dict(df):
            sd = status_default_dict()
            for row in df.itertuples():
                sd[row.task][row.modality].append(
                    {key: getattr(row, key) for key in columns}
                )
            return sd

        return get_status_dict(df)


class Record(RecordUnit, Scope):
    """
    A Record instance is used to manipulate a single record data.
    It's accessed from a [RecordCollection](record_collection.md).
    
    Note:
        Inherite [Scope](scope.md) methods.

    Example:

    ```python
    rec = ws.get("RECORD-ID") # ws is a Workspace instance
    ```
    """

    def logs(self) -> pd.DataFrame:
        """
        Returns:
            Logs history.
        """

        log_path = record_log_file_path(self.workspace.data_path, self.id)
        if log_path.exists():
            json = "[{}]".format(",".join(log_path.read_text().split("\n")[:-1]))
            df = pd.read_json(StringIO(json), orient="records")
            for key in ("task", "modality"):
                k = getattr(self, key)
                if k is not None:
                    df = df.query("{} == '{}'".format(key, k))
            return df.sort_values("time", ascending=False)
        else:
            return pd.DataFrame()

    def _count_detail_unit(self):

        return pd.DataFrame(
            [
                {
                    "record_id": self.id,
                    "task": self.task,
                    "modality": self.modality,
                    "count": self._unit_count_raw_data(),
                }
            ]
        )

    def _unit_count_raw_data(self):
        if not self.is_unit:
            raise NotUnitError
        first_step = self.steps.first()
        if first_step and first_step.script.data_exist.input:
            return 1
        return 0

    def status_dict(self) -> str:
        """
        Returns:
            Status in dict format.
        """

        columns = ("step_index", LOG_STEP_LABEL, LOG_STEP_STATUS)

        return self._structure_status(columns)

    def status_json(self) -> str:
        """
        Returns:
            Status in JSON format for web API.
        """

        return json.dumps(self.status_dict())



# coding=utf-8

import re
from pathlib import Path
from typing import Union, Optional
import toml
from sinagot.models import (
    RecordCollection,
    RecordCollectionUnit,
    Record,
    RecordUnit,
    Model,
)
from sinagot.config import ConfigurationError
from sinagot.logger import logger_factory
from sinagot.run_manager import RunManager

try:
    from sinagot.plugins.dask import DaskRunManager

    dask_enable = True
except ImportError:
    dask_enable = False


class Workspace(Model):
    """
    Workspace is the main class to handle dataset and scripts. It required a configuration file in `.toml` format.
    
    You can access to the [RecordCollection](record_collection.md) of the dataset with `.records` property.
    
    You can access to the [StepCollection](step_collection.md) of the scripts with `.steps` property.

    Usage:
        Create an instance :
        
            from sinagot import Workspace
            ws = Workspace(path/to/config.toml)
    """

    DEFAULT_CONFIG = {
        "path": {"dataset": "./dataset", "scripts": "./scripts"},
        "run": {"mode": "main_process"},
    }

    def __init__(
        self,
        config_path: Union[str, Path],
        data_path: Optional[Union[str, Path]] = None,
        scripts_path: Optional[Union[str, Path]] = None,
    ):
        """
        Arguments:
            config_path: path to the config file in toml.
            data_path: path to the folder of the workspace.
        """

        super().__init__(self)

        self._load_config(config_path)
        self._set_data_path(data_path)
        self._set_scripts_path(scripts_path)
        self._init_run_logger()
        self._init_run_manager()
        self._init_records()

    def _load_config(self, config_path):
        self._config_path = Path(config_path)
        if self._config_path.is_dir():
            self._config_path = self._config_path / "workspace.toml"
        config = toml.load(self._config_path)
        for section, values in self.DEFAULT_CONFIG.items():
            if section not in config:
                config[section] = values
        self.config = config
        """Config dictionnary from config file."""
        self.is_unit_mode = not any(
            scope in self.config for scope in ("tasks", "modalities")
        )

    def _set_data_path(self, data_path):
        if data_path is None:
            data_path = self.config["path"].get("dataset", self._config_path)
        self._data_path = self._resolve_path(data_path)

    @property
    def data_path(self) -> Path:
        """Path of data folder"""
        return self._data_path

    def _set_scripts_path(self, scripts_path):
        scripts_path = scripts_path or self.config.get("path", {}).get("scripts")
        if scripts_path:
            self._scripts_path = self._resolve_path(scripts_path)
        else:
            self._scripts_path = self._config_path.parent

    @property
    def scripts_path(self) -> Path:
        """Path of all the scripts folder"""
        return self._scripts_path

    def _init_run_logger(self):
        self.logger = logger_factory(self.config, self.is_unit_mode)
        """Logger of the workspace"""
        log_path = self.data_path / "LOG"
        if not log_path.exists():
            log_path.mkdir()

    def _init_run_manager(self):
        try:
            run_mode = self.config["run"]["mode"]
        except KeyError:
            run_mode = "main_process"
        if run_mode == "dask":
            if not dask_enable:
                raise ConfigurationError("Dask is not enable in this environment")
            run_manager = DaskRunManager
        else:
            run_manager = RunManager

        self._run_manager = run_manager(self)

    def _resolve_path(self, raw_path):
        """Resolve path from config from raw_path and self.config_path if raw path
        is relative to current dir i.e. begin with './' or '...'

        Args:
            raw_path (str or pathlib.Path): path from config to resolve

        Returns:
            pathlib.Path: absolute path
        """

        path = Path(raw_path)
        if re.match(r"(?:\.{1,2}\/.*|\.$)", str(raw_path)):
            path = Path(self._config_path.parent, raw_path).resolve()
        return path

    @property
    def records(self) -> RecordCollection:
        """RecordCollection of all records of the workspace"""
        return self._records

    @property
    def steps(self) -> "StepCollection":
        """StepCollecion for all steps of the workspace"""
        return self._records.steps

    def _init_records(self):

        if self.is_unit_mode:
            RecordCollectionUnit.__name__ = "RecordCollection"
            RecordUnit.__name__ = "Record"
            self._records = RecordCollectionUnit(self)
        else:
            self._records = RecordCollection(self)
            RecordCollection._set_subscopes(self)
            Record._set_subscopes(self)
            self._alias_subscopes()

    def _alias_subscopes(self):
        ALIASES = self.records._tasks + self.records._modalities
        for alias in ALIASES:
            setattr(self, alias, getattr(self.records, alias))

    def close(self):
        self._run_manager.close()



# coding=utf-8

import pandas as pd
from sinagot.models import Record, RecordCollection


class ProcessedRecord(Record):

    _DATA_TYPES = {
        "str": str,
        "float": float,
        "int": int,
    }

    _CONFIG_PROCESSED_LABEL = "processed"
    _DATA_GETTER_LABEL = "data_getter"

    @classmethod
    def _data_getters(cls, getter_label):
        GETTERS = {
            "single_data": cls._get_single_data,
            "dataframe_from_csv": cls._get_dataframe_from_csv,
            "series_from_csv": cls._get_series_from_csv,
        }
        return GETTERS[getter_label]

    def get_processed_data(self, data_key):
        params = self.workspace.config[self._CONFIG_PROCESSED_LABEL][data_key]
        path = self._get_raw_path(params)
        if not path.exists():
            return pd.DataFrame(columns=params.get("columns"))
        getter_label = params.get(self._DATA_GETTER_LABEL, "single_data")
        getter = self._data_getters(getter_label)
        return getter(self, path, params).set_index("record_id")

    def _get_raw_path(self, params):
        record = Record(
            self.workspace,
            record_id=self.id,
            task=params["task"],
            modality=params["modality"],
        )
        step = record.steps.get(params["step_label"])
        path = step.script.path.output
        path_label = params.get("path_label")
        if path_label:
            path = path[path_label]
        return path

    def _get_single_data(self, path, params):

        data = path.read_text()
        data_type = params.get("data_type")
        if data_type:
            data = self._DATA_TYPES[data_type](data)
        return pd.DataFrame(
            [
                {
                    "record_id": self.id,
                    "task": params["task"],
                    "modality": params["modality"],
                    params["step_label"]: data,
                }
            ]
        )

    def _get_dataframe_from_csv(self, path, params=None):
        data = pd.read_csv(path)
        data = self._add_record_id_to_dataframe(data)
        return data

    def _get_series_from_csv(self, path, params=None):
        data = pd.read_csv(path, header=None, index_col=0).transpose()
        columns = params.get("columns")
        if columns:
            data = data.reindex(columns=columns)
        data = self._add_record_id_to_dataframe(data)
        return data

    def _add_record_id_to_dataframe(self, data):
        data.insert(0, "record_id", self.id)
        return data


class ProcessedRecordCollection(RecordCollection):

    _record_class = ProcessedRecord

    def get_processed_data(self, **kwargs):
        return pd.concat([rec.get_processed_data(**kwargs) for rec in self.all()])


MODELS = {"record": ProcessedRecord, "record_collection": ProcessedRecordCollection}



from pathlib import Path
from dask import visualize
from dask.delayed import Delayed
from dask.distributed import LocalCluster, Client, fire_and_forget
from sinagot.utils import get_module, get_script
from sinagot.run_manager import RunManager
from sinagot.config import ConfigurationError
from sinagot.run_manager import run_step_factory


class DaskRunManager(RunManager):
    """Manage multiple run in sequential or parallel mode"""

    def __init__(self, workspace):
        super().__init__(workspace)
        self.dask_config = workspace.config.get("dask", {})
        self.init_scheduler()

    def init_scheduler(self):
        scheduler_type = self.dask_config.get("scheduler", {}).get("type", "processes")
        if scheduler_type == "distributed":
            cluster_config = self.dask_config.get("cluster", {})
            self.scheduler = DaskDistributedScheduler(cluster_config)
        else:
            self.scheduler = DaskLocalScheduler(scheduler_type)

    def visualize(self, records, **kwargs):
        graph = DaskGraph(self.workspace)
        graph.build(records)
        return visualize(graph.dsk, **kwargs)

    def _run(self, records, run_opts):
        for record in records:
            graph = DaskGraph(self.workspace)
            graph.build(record, run_opts)
            dl = Delayed(("record", record.id), graph.dsk)
            self.scheduler.compute(dl)


class DaskLocalScheduler:
    def __init__(self, scheduler):
        self.scheduler = scheduler

    def compute(self, operation):
        return operation.compute(scheduler=self.scheduler)

    def close(self):
        pass


class DaskDistributedScheduler:

    _cluster = None
    _client = None

    def __init__(self, cluster_config):
        self.cluster_config = cluster_config

    def compute(self, operation):
        fire_and_forget(self.client.compute(operation))

    @property
    def client(self):
        if not self._client:
            self._init_client()
        return self._client

    def _init_client(self):
        self._client = Client(self.cluster)

    @property
    def cluster(self):
        if not self._cluster:
            self._init_cluster()
        return self._cluster

    def _init_cluster(self):
        cluster_config = self.cluster_config
        address = cluster_config.pop("address", None)
        if address:
            self._cluster = address
        else:
            self._cluster = LocalCluster(**cluster_config)

    def close(self):
        if self._client:
            self._client.close()
        if self._cluster:
            self._cluster.close()


class DaskGraph:
    def __init__(self, workspace):
        self.workspace = workspace
        self.dsk = {}

    @staticmethod
    def group(*args):
        pass

    def build(self, record, run_opts=None):
        # for record in records:
        task_outs = []
        if record.task:
            tasks = [record]
        else:
            tasks = record.iter_tasks()
        for task in tasks:
            mod_outs = []
            if task.modality:
                modalities = [task]
            else:
                modalities = task.iter_modalities()
            for modality in modalities:
                step = None
                prev_step = "raw"
                params = {
                    "record_id": modality.id,
                    "task": modality.task,
                    "modality": modality.modality,
                }
                for step in modality.steps.scripts_names():
                    target = self.add_step(prev_step, step, **params, run_opts=run_opts)
                    prev_step = step
                if step:
                    mod_outs.extend(target)
            task_outs.append(
                self.add_edge(
                    mod_outs,
                    self.group,
                    ("task", "{record_id}-{task}".format(**params)),
                )
            )
        self.add_edge(task_outs, self.group, ("record", "{record_id}".format(**params)))

    def add_edge(self, sources, func, target):
        graph = self.dsk
        graph.update({target: (func, *sources)})
        self.dsk = graph
        return target

    def add_step(self, source, target, record_id, task, modality, run_opts):

        step_params = {
            "workspace": self.workspace,
            "record_id": record_id,
            "task": task,
            "modality": modality,
        }

        func = run_step_factory(**step_params, step_label=target, run_opts=run_opts)

        script = get_script(**step_params, step_label=target)
        path_in = script.path.input
        if isinstance(path_in, dict):
            path_ins = path_in.values()
        else:
            path_ins = (path_in,)
        keys_in = tuple((source, self.format_path(p)) for p in path_ins)
        path_out = script.path.output
        if isinstance(path_out, dict):
            out_keys = [self.format_path(out_value) for out_value in path_out.values()]
            key_out = (target, *out_keys)
            res = []
            for out_value in out_keys:
                res_item = (target, out_value)
                self.add_edge(
                    ((target, *out_keys),), self.split_out, res_item,
                )
                res.append(res_item)

        else:
            key_out = (target, self.format_path(path_out))
            res = (key_out,)
        self.add_edge(keys_in, func, key_out)
        return res

    def format_path(self, path):
        return str(path.relative_to(self.workspace.data_path))

    @staticmethod
    def split_out(*args):
        pass

    @staticmethod
    def get_path(script, direction):
        path = getattr(script.path, direction)
        if isinstance(path, Path):
            path = {"data": path}
        return path



class NotUnitError(Exception):
    """Rise when a model is not a unit"""


class NoModalityError(Exception):
    """Rise when a model has no specific modality"""


class NotFoundError(Exception):
    """Rise when finding method return nothing"""



from importlib import import_module, util as importutil
from pathlib import Path

LOG_ORIGIN = "log_from"
LOG_STEP_LABEL = "step_label"
LOG_RECORD_ID = "record_id"
LOG_STEP_STATUS = "step_status"


class StepStatus:

    INIT = 0
    DATA_READY = 10
    PROCESSING = 20
    DONE = 30
    ERROR = 40
    DATA_NOT_AVAILABLE = 41


def record_log_file_path(data_path, rec_id):
    return data_path / "LOG" / "{}.log".format(rec_id)


def get_module(workspace, class_name, *args):
    """Import a module from scripts folder"""

    path = Path(workspace._scripts_path, *args)
    spec = importutil.spec_from_file_location(
        ".".join(args[-3:]), path.with_suffix(".py")
    )
    module = importutil.module_from_spec(spec)
    spec.loader.exec_module(module)
    return getattr(module, class_name)


def get_plugin_modules(plugin_model):

    plugin = import_module("sinagot.plugins.models." + plugin_model)
    return getattr(plugin, "MODELS")


def get_script(workspace, record_id, task, modality, step_label):
    if modality:
        script_class = get_module(workspace, "Script", modality, step_label)
    else:
        script_class = get_module(workspace, "Script", step_label)
    return script_class(
        data_path=workspace.data_path,
        record_id=record_id,
        task=task,
        logger_namespace=workspace.logger.name,
    )



# coding=utf-8

from .model import Model
from .step import Step, ScopedStep
from .step_collection import (
    StepCollectionUnit,
    StepCollection,
    ModelWithStepCollection,
)
from .scope import Scope
from .record import RecordUnit, Record
from .record_collection import RecordCollectionUnit, RecordCollection

Subset = RecordCollection
from .workspace import Workspace, ConfigurationError


__all__ = [
    "Model",
    "Scope",
    "Workspace",
    "ConfigurationError",
    "Subset",
    "RecordCollectionUnit",
    "RecordCollection",
    "Record",
    "RecordUnit",
    "Step",
    "ScopedStep",
    "StepCollectionUnit",
    "StepCollection",
    "ModelWithStepCollection",
]



# coding=utf-8

from typing import Optional, Union
import pandas as pd
from sinagot.models import Model, Step, ScopedStep
from sinagot.utils import LOG_STEP_LABEL, LOG_STEP_STATUS
from sinagot.models.exceptions import NotFoundError, NotUnitError


class ModelWithStepCollection(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.steps: StepCollection = self.set_step_collection()
        """Collection of steps."""

    def set_step_collection(self):
        if self.workspace.is_unit_mode:
            return StepCollectionUnit(self)
        return StepCollection(self)


class StepCollectionUnit(Model):
    """Manage the collection of all steps."""

    _MODEL_TYPE = "step_collection"
    _STEP_CLASS = Step
    task = None
    modality = None

    def __init__(self, model):
        """
        Params:
            model (instance): Parent model of the collection.
        """
        super().__init__(model.workspace)
        self.model = model

    def scripts_names(self):
        return self._scripts_config

    @property
    def _scripts_config(self):
        return self.workspace.config.get("steps", {}).get("scripts", [])

    def get(self, script_name: str) -> Step:
        """find a step by script name.

        Params:
            script_name: script label to find.

        Returns:
            `Step` instance.
        """
        return self._get(script_name)

    def _get(self, script_name):
        return self._STEP_CLASS(script=script_name, model=self.model)

    def first(self) -> Step:
        """Get the first step.

        Returns:
            `Step` instance.
        """
        return self._first()

    def _first(self):
        names = self.scripts_names()
        if not names:
            return None
        return self.get(names[0])

    def count(self) -> int:
        """

        Returns:
            number of steps.
        """
        return self._count()

    def _count(self):
        return len(self.scripts_names())

    def run(
        self,
        step_label: Optional[str] = None,
        force: Optional[bool] = False,
        ignore_missing: Optional[bool] = False,
        debug: Optional[bool] = False,
    ):
        """Run all steps of the scope.
        
        Args:
            step_label: if not `None`, run only for the step with this label.
            force: Force run and overwrites result file(s) if already exist(s).
            debug: If False, log scripts errors and not raise them.
        """

        workspace = self.workspace
        return workspace._run_manager.run(
            self.model,
            step_label=step_label,
            force=force,
            ignore_missing=ignore_missing,
            debug=debug,
        )

    def status(self) -> pd.DataFrame:
        """Get status of all steps

        Returns:
            A pandas DataFrame with a row for each step.
        """
        if self.model._MODEL_TYPE == "record":
            return self._record_status()
        elif self.model.count() == 0:
            return pd.DataFrame([])
        else:
            return pd.concat([rec.steps.status() for rec in self.model.all()])

    def _record_status(self):
        if self.count():
            return pd.DataFrame(
                [
                    self._record_status_unit(index, step)
                    for index, step in zip(range(self.count()), self._modality_all(),)
                ]
            )
        else:
            return None

    def _record_status_unit(self, index, step):
        model = self.model
        if not model.is_unit:
            raise NotUnitError
        return {
            "record_id": self.model.id,
            "task": self.task,
            "modality": self.modality,
            "step_index": index + 1,
            LOG_STEP_LABEL: step.label,
            LOG_STEP_STATUS: step.status(),
        }

    def _modality_all(self):
        for script_name in self.scripts_names():
            yield self.get(script_name)


class StepCollection(StepCollectionUnit):
    """Manage the collection of all steps of a scope."""

    _STEP_CLASS = ScopedStep

    def __init__(self, model):
        """
        Params:
            model (instance): Parent model of the collection.
        """
        super().__init__(model)
        self.task = self.model.task
        self.modality = self.model.modality

    def get(self, script_name: str) -> Union[Step, dict]:
        """find a step by script name.

        Params:
            script_name: script label to find.

        Returns:
            `Step` if model has modality, `dict` with `Step` or `None` as value for each modality either.
        """
        return self._modality_case_response("_get", script_name)

    def first(self) -> Union[Step, dict]:
        """Get the first step.

        Returns:
            `Step` if model has modality, `dict` with first step as value for each modality either.
        """
        return self._modality_case_response("_first")

    def count(self) -> Union[int, dict]:
        """Get the number of steps.

        Returns:
            `int` if model has modality, `dict` with step count as value for each modality either.
        """
        return self._modality_case_response("_count")

    def _modality_case_response(self, method, *args):
        if self.model.modality:
            return getattr(self, method)(*args)
        else:
            return {
                modality.modality: modality.steps._modality_case_default(method, *args)
                for modality in self.model.iter_modalities()
            }

    def _modality_case_default(self, method, *args):
        try:
            return getattr(self, method)(*args)
        except NotFoundError:
            return None

    def _record_status(self):
        if self.count():
            return pd.DataFrame(
                [
                    unit.steps._record_status_unit(index, step)
                    for unit in self.model.iter_units()
                    for index, step in zip(
                        range(unit.steps.count()), unit.steps._modality_all()
                    )
                ]
            )
        else:
            return None

    def scripts_names(self):
        if self.modality:
            return self._modality_scripts_names()
        else:
            return {
                modality.modality: modality.steps._modality_scripts_names()
                for modality in self.model.iter_modalities()
            }

    def _modality_scripts_names(self):
        mod_config = self.workspace.config["modalities"].get(self.modality, {})
        shared_names = mod_config.get("scripts", [])
        task_scripts = mod_config.get("task_scripts", {})
        if self.task:
            task_names = task_scripts.get(self.task, [])
        else:
            task_names = [
                script for scripts in task_scripts.values() for script in scripts
            ]
        return shared_names + task_names



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sinagot', 'sinagot.models', 'sinagot.plugins', 'sinagot.plugins.models']

package_data = \
{'': ['*']}

install_requires = \
['json-log-formatter>=0.3.0,<0.4.0',
 'pandas>=1.0.2,<2.0.0',
 'toml>=0.10.0,<0.11.0']

extras_require = \
{'dask': ['dask>=2.12.0,<3.0.0', 'distributed>=2.12.0,<3.0.0']}

setup_kwargs = {
    'name': 'sinagot',
    'version': '0.2.4',
    'description': 'Python library to batch scripts on a file-system workspace.',
    'long_description': '# Sinagot\n\nSinagot is a Python library to batch multiple **scripts** on a file-system **dataset** with a simple API.\nParallelization of data processing is made possible by [Dask.distributed](https://distributed.dask.org/en/latest/). \n\n## Installation\n\nSinagot is available on PyPi:\n\n```bash\npip install sinagot\n```\n\n## Full Documentation\n\n<https://sinagot.readthedocs.io>\n\n## Concept\n\nSinagot main class is build around the `sinagot.Workspace` class. To create an instance, you must provide 3 pathes to :\n\n- A configuration file in `.toml` format.\n- A data folder.\n- A scripts fodler.\n\n<img src="https://github.com/YannBeauxis/sinagot/raw/master/docs/workspace_structure.png" width="200">\n\nDataset is structured as a collection of **records**. A record is identified by an unique ID but many files can be generated for a single record. Those files are processed with **scripts** which generate other files as results.\n\n## Basic example \n\n### Harbor workspace\n\nYou can find in "example" folder of the git the [harbor](https://github.com/YannBeauxis/sinagot/tree/master/example/harbor) workspace that has a record per day of a harbor occupancy. In this example, a record is created each day to count the boats that occupy the harbor. The record ID include a timestamp for the day of recording.\n\nIn Unix environment, you can that type this to get the workspace :\n\n```bash\nwget -qO- https://github.com/YannBeauxis/sinagot/raw/master/example/harbor.tar.gz | tar xvz\n```\n\nTo create the workspace instance :\n\n```python\n>>> from sinagot import Workspace\n>>> ws = Workspace(\'/path/to/harbor/workspace/folder\')\n>>> ws\n<Workspace instance>\n```\n\n### Explore records\n\nYou can list all records ids:\n\n```python\n>>> list(ws.records.iter_ids())\n[\'REC-20200602\', \'REC-20200603\', \'REC-20200601\']\n```\n\nCreate a `Record` instance. For a specific record:\n\n```python\n>>> ws.records.get(\'REC-20200603\')\n<Record instance | id: REC-20200603>\n```\n\nOr the first record found:\n\n```python\n>>> ws.records.first()\n<Record instance | id: REC-20200602>\n```\n\n> Records are not sort by their ids.\n\n### run scripts\n\nYou can run all scripts for each record of the dataset:\n\n```python\n>>> ws.steps.run()\nREC-20200602 | 2020-08-20 11:19:11,530 |\xa0count : Init run\nREC-20200602 | 2020-08-20 11:19:11,531 |\xa0count : Processing run\nREC-20200602 | 2020-08-20 11:19:11,556 |\xa0count : Run finished\n...\nREC-20200601 | 2020-08-20 11:19:11,625 |\xa0mean : Init run\nREC-20200601 | 2020-08-20 11:19:11,626 |\xa0mean : Processing run\nREC-20200601 | 2020-08-20 11:19:11,634 |\xa0mean : Run finished\n\n```\n\nOr for a single record:\n\n```python\n>>> ws.records.get(\'REC-20200603\').steps.run()\nREC-20200603 | 2020-08-20 11:28:32,588 |\xa0count : Init run\nREC-20200603 | 2020-08-20 11:28:32,590 |\xa0count : Processing run\nREC-20200603 | 2020-08-20 11:28:32,616 |\xa0count : Run finished\nREC-20200603 | 2020-08-20 11:28:32,619 |\xa0mean : Init run\nREC-20200603 | 2020-08-20 11:28:32,621 |\xa0mean : Processing run\nREC-20200603 | 2020-08-20 11:28:32,637 |\xa0mean : Run finished\n```\n\n## More complex dataset\n\nYou can handle more complexity of dataset structure with **task** and **modality** concepts. During a recording session for a single record, data can be generate for differents task and each task can generate different kind of data called **modality**. \n\n### SoNeTAA usecase\n\nThe idea of Sinagot emerged for the data management of an EEG platform called SoNeTAA :\nhttps://research.pasteur.fr/en/project/sonetaa/ .\n\nFor documentation purpose SoNeTAA workspace structure will be used as example. \n\nOn SoNeTAA, a record with an ID with timestamp info in this format `REC-[YYMMDD]-[A-Z]`, \nfor example `"REC-200331-A"`. \n\nFor a record, 3 tasks are performed: \n\n* "RS" for Resting State\n* "MMN" for MisMatch Negativity\n* "HDC" for Human Dynamic Clamp.\n\n3 modalities handle data depending of the tasks\n* For each tasks, "EEG" modality create data from ElectroEncephalogram .\n* A "behavior" modality create date only for HDC task.\n* A "clinical" modality handle data used for every task.\n\n### Explore by task or modality\n\nEach record collection or single record has **subscopes** corresponding to their tasks and modalities accessible as attribute.\n\nFor example to select only the task RS of the dataset:\n\n```python\n>>> ws.RS\n<RecordCollection instance | task: RS, modality: None>\n```\n\n> A dataset subscope is a **RecordCollection**.\n\nOr the EEG modality of a record:\n\n```python\n>>> rec.EEG\n<Record instance | id: REC-200331-A, task: None, modality: EEG>\n```\n\nYou can select a specific couple of task and modality (called **unit**):\n\n```python\n>>> ws.RS.EEG\n<RecordCollection instance | task: RS, modality: EEG>\n>>> ws.EEG.RS\n<RecordCollection instance | task: RS, modality: EEG>\n```\n',
    'author': 'Yann Beauxis',
    'author_email': 'dev@yannbeauxis.net',
    'maintainer': 'Yann Beauxis',
    'maintainer_email': 'dev@yannbeauxis.net',
    'url': 'https://github.com/YannBeauxis/sinagot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# coding=utf-8
"""Common functions used for script"""

import logging
from pathlib import Path
from typing import Optional
from collections import namedtuple
from json_log_formatter import JSONFormatter
from sinagot.utils import (
    record_log_file_path,
    LOG_ORIGIN,
    LOG_STEP_LABEL,
    LOG_RECORD_ID,
    LOG_STEP_STATUS,
    StepStatus,
)


class ScriptTemplate:
    """
    Template class for scripts. You must override `PATH_IN`, `PATH_OUT` and `run()` to use it.

    Usage:
        From harbor example :

            import pandas as pd
            from sinagot import ScriptTemplate

            class Script(ScriptTemplate):

                PATH_IN = ("raw", "{id}-raw.csv")
                PATH_OUT = ("computed", "{id}-count.csv")

                def run(self):
                    df = pd.read_csv(self.path.input)
                    df = df.groupby("country").count()
                    df.to_csv(self.path.output)
    """

    _PATH_LABELS = ["input", "output"]
    PATH_IN = ("FOLDER_IN", "{id}-{task}.in")
    """tuple or dict of tuples to specify input path pattern"""
    PATH_OUT = ("FOLDER_OUT", "{id}-{task}.out")
    PATH_CONTROL = ("FOLDER_CONTROL", "{id}-{task}.control")
    """tuple or dict of tuples to specify output path pattern"""

    debug = False
    _logger_file_handler = None
    _logger = None
    logger = None
    modality = None
    """The modality label"""

    def __init__(
        self,
        data_path: str,
        record_id: str,
        task: Optional[str] = None,
        opts={},
        logger_namespace: Optional[str] = None,
    ):

        self.status = StepStatus.INIT
        """
        Status of the script during its run. Valuse are defined in sinagot.utils.StepStatus class"""
        self.data_path = data_path
        """The path to dataset"""
        self.id = record_id
        """The record ID"""
        self.task = task
        """The modality label"""
        module_split = self.__class__.__module__.split(".")
        if len(module_split) > 1:
            self.modality = module_split[-2]
        self.label = module_split[-1]
        "The step label"
        self.opts = opts
        "Optional dict"
        self._logger_namespace = logger_namespace

        self._set_path()

    def _init_logger(self, debug=False):
        file_handler_path = record_log_file_path(self.data_path, self.id)
        file_handler = logging.FileHandler(file_handler_path)
        file_formatter = JSONFormatter()
        file_handler.setFormatter(file_formatter)
        file_handler.addFilter(self._log_filter_factory())
        self._logger_file_handler = file_handler
        logger_ = logging.getLogger(self._logger_namespace)
        if debug:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logger_.setLevel(log_level)
        self._logger = logger_
        logger = logging.LoggerAdapter(logger_, self._log_extra(StepStatus.PROCESSING))
        self.logger = logger

    def _log_filter_factory(self):
        record_id = self.id

        def record_filter(record):
            return record.__dict__[LOG_RECORD_ID] == record_id

        return record_filter

    def _log_extra(self, status):
        return {
            LOG_ORIGIN: "script",
            LOG_RECORD_ID: self.id,
            "scope": "record",
            "task": self.task,
            "modality": self.modality,
            LOG_STEP_LABEL: self.label,
            LOG_STEP_STATUS: status,
        }

    def _set_path(self):

        ScriptPath = namedtuple("IOPath", self._PATH_LABELS + ["control"])
        self.path = ScriptPath(
            input=self._get_path(self.PATH_IN),
            output=self._get_path(self.PATH_OUT),
            control=self._get_path(self.PATH_CONTROL),
        )

    def _get_path(self, raw_path):

        if isinstance(raw_path, tuple):
            return Path(self.data_path, *self._str_path(raw_path))
        if isinstance(raw_path, dict):
            return {
                label: Path(self.data_path, *self._str_path(rp))
                for label, rp in raw_path.items()
            }

    def _str_path(self, raw_path_unit: tuple):
        return (
            rp.format(id=self.id, task=self.task, **self.opts) for rp in raw_path_unit
        )

    def _run(
        self, force: bool = False, ignore_missing: bool = False, debug: bool = False
    ):
        """Required run function. Called by step model :code:`run()` method."""
        self.debug = debug
        self._init_logger(debug)
        self._logger.addHandler(self._logger_file_handler)
        self._log_status("Init run", StepStatus.INIT)
        if not all(self.data_exist.input.values()) and not ignore_missing:
            missing = [key for key, exist in self.data_exist.input.items() if not exist]
            self._log_status(
                "Input data  %s not available", StepStatus.DATA_NOT_AVAILABLE, missing
            )
        elif not self._has_to_run(force):
            self._log_status("Run already processed", StepStatus.DONE)
        else:
            self.status = StepStatus.PROCESSING
            self._log_status("Processing run", StepStatus.PROCESSING)
            try:
                self._mkdir_output()
                self.run()
                self._log_status("Run finished", StepStatus.DONE)
            except Exception as ex:
                if debug:
                    raise ex
                self._log_status(ex, StepStatus.ERROR)

        self._logger.removeHandler(self._logger_file_handler)

        return True

    def _log_status(self, message, status, *args):
        self._logger.info(message, *args, extra=self._log_extra(status))

    def run(self):
        """Main method called during step run."""
        self.logger.info("Running ...")

    @property
    def data_exist(self):
        """Get named tuple of boolean to check if input and output data exist."""
        DataStatus = namedtuple("IOPath", self._PATH_LABELS)
        return DataStatus(*(self._path_exist(target) for target in self._PATH_LABELS))

    def _path_exist(self, target):
        return {path: path.exists() for path in self._iter_paths(target)}

    def _mkdir_output(self):
        for path in self._iter_paths("output"):
            if not path.parent.exists():
                path.parent.mkdir(parents=True)

    def _iter_paths(self, target):
        if target not in ("input", "output"):
            raise AttributeError("target not in ('input', 'output')")
        paths = getattr(self.path, target)
        if isinstance(paths, Path):
            return (paths,)
        return paths.values()

    def _has_to_run(self, force):
        if force:
            return True
        for path in self._iter_paths("output"):
            if not path.exists():
                return True
        return False



# coding=utf-8

from typing import Optional
from pathlib import Path
import re
from typing import Generator
import pandas as pd
from sinagot.utils import get_module
from sinagot.models import ModelWithStepCollection, Scope, Record, RecordUnit


class RecordCollectionUnit(ModelWithStepCollection):

    _MODEL_TYPE = "record_collection"
    _record_class = RecordUnit
    task = None

    def ids(self) -> list:
        """List of all records ids
    
        Returns:
            Record ids"""
        return list(self.iter_ids())

    def iter_ids(self):
        """Generator all record ids within the record_collection.
        
        Returns:
            Record ID.
        """
        self._ids = []

        path_raw = self._iter_path()

        if path_raw:
            if isinstance(path_raw, dict):
                path_tuple = path_raw.values()[0]
            else:
                path_tuple = path_raw

            root_path = Path(self.workspace.data_path)
            raw_pattern = Path(*path_tuple)
            glob_pattern = str(self._glob_pattern(raw_pattern))
            re_pattern = re.compile(str(root_path / self._re_pattern(raw_pattern)))

            for path in root_path.glob(glob_pattern):
                record_id = self._evaluate_path(re_pattern, path)
                if record_id:
                    yield record_id

    def _iter_path(self):
        first_script = self.steps.first()
        if first_script:
            return first_script.script.PATH_IN

    def _evaluate_path(self, re_pattern, path):
        m = re_pattern.search(str(path))
        if m and (len(m.groups()) > 0):
            record_id = m.group(1)
            if self._is_new_id(record_id):
                return record_id

    def _is_new_id(self, record_id):
        ids = self._ids
        if record_id in ids:
            return False
        ids.append(record_id)
        self._ids = ids
        return True

    def _glob_pattern(self, raw_pattern):
        return str(raw_pattern).format(id="*", task="*",)

    def _re_pattern(self, raw_pattern):
        return str(raw_pattern).format(
            id="({})".format(self.config["records"]["id_pattern"]),
            task="(?:{})".format(self.task),
        )

    def all(self) -> Generator["Record", None, None]:
        """Generate all records instances of the record_collection.
        
        Returns:
            sinagot.models.Record: Record instance
        """

        for record_id in self.iter_ids():
            yield self.get(record_id)

    def first(self) -> "Record":
        """
        Get the first record found.
        
        Returns:
            Record instance
        """

        for record_id in self.iter_ids():
            return self.get(record_id)

    def head(self, count=5) -> "Record":
        """
        Get the first record found.
        
        Returns:
            Record instance
        """

        idx = 0
        for record_id in self.iter_ids():
            yield self.get(record_id)
            idx += 1
            if idx > count:
                break

    def get(self, record_id: str) -> "Record":
        """
        Get record by ID.

        Args:
            record_id: record ID

        Returns:
            Record instance
        """

        return self._record_class(workspace=self.workspace, record_id=record_id,)

    def has(self, record_id: str) -> bool:
        """
        Check existance of a record in the record_collection.
        
        Args:
            record_id: Id of the record.
        
        Returns:
            True if record exists.
        """

        for id_ in self.iter_ids():
            if id_ == record_id:
                return True
        return False

    def count(self) -> int:
        """
        Count all records.

        Returns:
            Number of records.
        """

        return sum(1 for rec in self.iter_ids())


class RecordCollection(RecordCollectionUnit, Scope):
    """
    A RecordCollection is used to manipulate a collection of [Record](record.md).

    Note:
        Inherite [Scope](scope.md) methods.

    Example:

    ```python
    # access record_collection of all EEG records (EEG is a task)
    sub = ws.EEG  # ws is a Workspace instance
    ```
    """

    _record_class = Record

    def __init__(self, *args, **kwargs):
        self._subscope_class = self.__class__
        super().__init__(*args, **kwargs)
        modality = self.modality
        # Check for custom record class
        if modality is not None:
            try:
                class_name = self.config["modalities"][modality]["models"]["record"]
                self._record_class = get_module(
                    self.workspace, class_name, modality, "models", "record"
                )
            except KeyError:
                pass
        self._ids = []

    def iter_ids(self) -> Generator[str, None, None]:
        """Generator all record ids within the record_collection.
        
        Returns:
            Record ID.
        """
        # TODO: use workspace cache for ids
        if self.is_unit:
            for record_id in super().iter_ids():
                yield record_id
        else:
            units = self.iter_units()
            self._ids = []
            for unit in units:
                for record_id in unit.iter_ids():
                    if self._is_new_id(record_id):
                        yield record_id

    def get(self, record_id: str) -> "Record":
        """
        Get record by ID.

        Args:
            record_id: record ID

        Returns:
            Record instance
        """

        return self._record_class(
            workspace=self.workspace,
            record_id=record_id,
            task=self.task,
            modality=self.modality,
        )

    def _count_detail_unit(self, with_record_id=False):

        df = pd.DataFrame(
            [
                {
                    "record_id": record_id,
                    "task": self.task,
                    "modality": self.modality,
                    "count": 1,
                }
                for record_id in self.iter_ids()
            ]
        )
        if df.empty or with_record_id:
            return df
        else:
            return df.groupby(["task", "modality"]).sum().reset_index()

    # TODO: To test
    def logs(self) -> pd.DataFrame:
        """
        Returns:
            Logs history.
        """
        return pd.concat([rec.logs() for rec in self.all()])
